#include "module.h"
#include "configstruct.h"

/*
  To add a bang command that applies to a window:

    - Add a DECL_BANG1 line for it at the end of the WindowBangs
      class.  The name as a parameter should not include the
      ModuleBangPrefix.  (eg, Move for !ModuleMove)

    - Add a DEF_BANG line and function body somewhere before
      the definition of the Bangs list.

      The function body is a member of the WindowBangs class, and
      can use any of its member functions.  You should NOT call
      other (local) bang commands from here.

      Available member variables:
        w     = pointer to the WndInfo struct for the window
                For multiple windows, the function will be called
                once for each window with w set appropriately.
        args  = arguments for bang command (after removing any
                window names)  Do not modify this.
        caller = caller of bang command
        one, two, three, tokens
              = variables set up for CommandTokenize
                Do not change the pointers in tokens.

    - Add a DECL_BANG2 line in the Bangs list at the end.


  To add a bang command that applies to the entire module:

    - Add a standard bang command definition and function body:
      void Bang<CommandName>(HWND,LPCSTR args)
      The function name should be the name of the bang command
      without ModuleBangPrefix and with "Bang" on the front.
      (eg, BangMove for !ModuleMove)

    - Add a DECL_BANG2 line in the Bangs list at the end.
      The name as a parameter should be the name of the bang
      command without ModuleBangPrefix.


  Once this is done, the bang command will automatically be
  registerred and unregisterred when the module is loaded or
  unloaded.  Window-specific bangs will automatically be
  applied to whichever window or windows are specified on the
  command line, or to the main window for a single-window
  module.

*/

//const char *togglestrings[] = { "toggle", "on", "yes", "true" };

#define DECL_BANG1(n) void Bang##n(void);
#define DECL_BANG2(n) { #n, &Bang##n },
/* might want to use BangCommandEx...  would be nice if we could associate an index with them */ 
#define DEF_BANG(n) void Bang##n(HWND caller,LPCSTR args) \
  { TRACE_MSG(("!"#n" %1",args)); ExecBang(&WindowBangs::Bang##n,caller,args); } \
  void WindowBangs::Bang##n(void)
#define EXEC_BANG(n) ExecBang(&WindowBangs::Bang##n,caller,args);

#undef MAX_LINE_LENGTH
#define MAX_LINE_LENGTH 256

class WindowBangs {
public:
  WndInfo *w;
  HWND caller;
  LPCSTR args;

  DECL_BANG1(Move)
  DECL_BANG1(Size)
  DECL_BANG1(OnTop)
  DECL_BANG1(Alpha)
  DECL_BANG1(Show)
#ifndef NO_SCRIPT_ACTIONS
  DECL_BANG1(Exec)
  DECL_BANG1(Action)
#endif
  DECL_BANG1(LoadBoxHook)
#ifndef NO_SKIN
  DECL_BANG1(Bitmap)
  DECL_BANG1(Edge)
  DECL_BANG1(Tile)
  DECL_BANG1(Color)
#endif
/* insert new DECL_BANG1 lines here */ 
};

#ifdef SINGLE_WINDOW
void ExecBang(void (WindowBangs::*cmd)(void),HWND caller,LPCSTR args)
{
  WindowBangs wb;
  ASSERT_EXECABLE(wb.*cmd);

  wb.w = &Module.MainWnd;
  wb.args = args;
  wb.caller = caller;
  (wb.*cmd)();
}
#else
void ExecBang(void (WindowBangs::*cmd)(void),HWND caller,LPCSTR args)
{
  char one[MAX_LINE_LENGTH];
  char *ptr,*end;
  int group;
  WindowBangs wb;
  WndInfo *pWnd;

// apparently wb.*cmd has no type recognizable to MSVC
// casting to void * first only results in an INTERNAL COMPILER ERROR
//  ASSERT_EXECABLE(wb.*cmd);

  wb.caller = caller;

  if (!GetToken(args,one,&wb.args,TRUE))
    return;

  /* window . means exec for entire module */ 
  ptr = one;
  while (*ptr==' ' || *ptr=='\t') ++ptr;
  if (*ptr=='.')
  {
    wb.w = NULL;
    (wb.*cmd)();
    return;
  }

  /* else exec for each window */ 
  end = ptr;
  while (*end)
  {
    if (*end==' ' || *end=='\t')
    {
      *end = 0;
      group = atoi(ptr);

      pWnd = Module.pWnds;
      while (pWnd)
      {
        if (group?pWnd->Group==group:match(ptr,pWnd->Name))
        {
          TRACE_MSG(("ExecBang for %1",pWnd->Name));
          wb.w = pWnd;
          (wb.*cmd)();
        }

        pWnd = pWnd->pNext;
      }

      ptr = ++end;
      while (*ptr==' ' || *ptr=='\t') ++ptr;
      end = ptr;
    }
    ++end;
  }

  if (*ptr)
  {
    group = atoi(ptr);

    pWnd = Module.pWnds;
    while (pWnd)
    {
      if (group?pWnd->Group==group:match(ptr,pWnd->Name))
      {
        TRACE_MSG(("ExecBang for %1",pWnd->Name));
        wb.w = pWnd;
        (wb.*cmd)();
      }

      pWnd = pWnd->pNext;
    }
  }
}
#endif /* SINGLE_WINDOW */ 


// !Move [monitor] <x> <y> [<width> <height>]
DEF_BANG(Move)
{
  if (!w) return;

  CallParser(NULL,args,&w->Pos,false);
  UpdateWndPos(w);
}

// !Size <width> <height>
DEF_BANG(Size)
{
  if (!w) return;

  CallParser(Size_ConfigItemPtr,NULL,args,&w->Pos,false);
  UpdateWndPos(w);
}

// !OnTop [bool]
DEF_BANG(OnTop)
{
  if (!w) return;

  CallParser(WndOnTop_ConfigItemPtr,NULL,args,w,false);
	UpdateWndOnTop(w);
}

// !Alpha <amount>
DEF_BANG(Alpha)
{
  if (!w) return;

  CallParser(Alpha_ConfigItemPtr,NULL,args,w,false);
	RedrawWnd(w);
}

// !Show [bool]
DEF_BANG(Show)
{
  if (!w) return;

  CallParser(WndShow_ConfigItemPtr,NULL,args,w,false);
	UpdateWndHidden(w);
}

#ifndef NO_SCRIPT_ACTIONS
DEF_BANG(Exec)
{
  ExecActionString(caller, w, args, NULL);
}
#endif

void BangLoadBoxHook(HWND, LPCSTR args)
{
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH],three[MAX_LINE_LENGTH];
	char four[MAX_LINE_LENGTH];
	char *tokens[3] = { one, two, three },*p;
	char name[MAX_LINE_LENGTH];
	char *lptr,*rptr;
	const char *rest;
	int i;
	FILE *f=NULL;
	WndInfo *w;
	Config c;

	rest = args;
	if (args && *args)
	{
		*name = 0;
		while (GetToken(rest,tokens[0],&rest,FALSE))
		{
			p = tokens[0];
			tokens[0] = tokens[1];
			tokens[1] = tokens[2];
			tokens[2] = p;
			lstrcpyn(four,tokens[0],5+1);
			if (lstrcmpi(four,"NAME=")==0)
			{
				lstrcpy(name,tokens[0]+5);
			}
		}
		/* tokens[2] = window handle */ 
		i = atoi(tokens[2]);
		if (!IsWindow((HWND)i)) 
		{
			MessageBox(NULL,"Invalid window handle in box hook.",ModuleName,MB_OK);
			return;
		}
		w = new WndInfo;
		ZeroMemory(w,sizeof(WndInfo));
		w->hParentWnd = (HWND)i;
		/* tokens[1] = box file */ 
		if (*tokens[1]==0 || (f = LCOpen(tokens[1])))
		{
			if (*name==0)
			{
				lstrcpy(name,ModuleRCPrefix);

				/* get name from box file name */ 
				if (*tokens[1])
				{
					lptr = NULL;
					rptr = tokens[1];
					while (*rptr!=0)
					{
						if (*rptr=='\\' || *rptr=='/') lptr = rptr+1;
						++rptr;
					}
					if (lptr)
					{
						lstrcpy(name,lptr);
						rptr = name;
						while (*rptr && *rptr!='.') ++rptr;
						*rptr = 0;
					}
				}
			}
		}
		w->Name = new char[lstrlen(name)+1];
		lstrcpy(w->Name,name);
		ZeroMemory(&c,sizeof(Config));
    ConfigState cs(&c,name,lstrlen(name));
    CallParser(&c,NULL,w,true);
    /* can't have layered in a box for some reason */ 
    w->bTrueAlpha = false;
		if (f) LCClose(f);
		AddWindow(w);
	}
}

/* insert new bang command definitions here */ 

#ifndef NO_SCRIPT_ACTIONS
DEF_BANG(Action)
{
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH];
	char *tokens[1] = { one };
  int iAction;

  TRACE(args);

	*one = *two = 0;
	LCTokenize(args,tokens,1,two);

  /* determine which action we're setting         */ 
  iAction = LSStringEnum(one,reinterpret_cast<const char **>(ScriptActionStrings),
    sizeof(*ScriptActionStrings));
  if (iAction<0) return;

  SetAction(w,iAction,two);
}
#endif /* NO_SCRIPT_ACTIONS */ 

#ifndef NO_SKIN
DEF_BANG(Bitmap)
{
	char one[MAX_LINE_LENGTH];

  if (!w) return;

  HBITMAP hbm;
  int cx,cy;
  BOOL HasAlpha;

  if (*args==':')
  {
    HDC a,b;
    HGDIOBJ oa,ob;
    const char *ptr;

	  GetToken(args+1,one,&ptr,FALSE);
    hbm = (HBITMAP)atoi(one);

    if (hbm)
    {
      /* get clip area */ 
      GetLSBitmapSize(hbm,&cx,&cy);
      RECT r = {0,0,cx,cy};
      CallParser(RECT_ConfigItemPtr,NULL,ptr,&r,true);

      if (r.top==-1) r.top=r.left=0;
      if (r.right==-1) r.right=cx-r.left;
      if (r.bottom==-1) r.bottom=cy-r.top;

      cx = r.right;
      cy = r.bottom;

      /* copy the bitmap */ 
      a = CreateCompatibleDC(NULL);
      b = CreateCompatibleDC(NULL);
      oa = SelectObject(a,hbm);
      ob = SelectObject(b,CreateCompatibleBitmap(a,cx,cy));
      BitBlt(b,0,0,cx,cy,a,r.left,r.top,SRCCOPY);
      SelectObject(a,oa);
      hbm = (HBITMAP)SelectObject(b,ob);
      DeleteDC(a);
      DeleteDC(b);

      HasAlpha = TRUE;
    }
    else if (*one=='.')
    {
      DeleteObject(w->emSkin.Bitmap);
      w->emSkin.Bitmap = hbm;
      RedrawWnd(w);
    }
  }
  else
  {
    hbm = LoadLSImage(args,NULL);
    GetLSBitmapSize(hbm,&cx,&cy);
    const char *p = strrchr(args,'.');
    if (p && stricmp(p,".png")==0)
      /* should make sure it's 32-bit... */ 
      HasAlpha = TRUE;
    else
      HasAlpha = FALSE;
  }

  if (hbm)
  {
    DeleteObject(w->emSkin.Bitmap);
    w->emSkin.Bitmap = hbm;
    w->emSkin.Size.cx = cx;
    w->emSkin.Size.cy = cy;
    w->emSkin.HasAlpha = !!HasAlpha;
    RedrawWnd(w);
  }
}

DEF_BANG(Edge)
{
  if (!w) return;

  CallParser(RECT_ConfigItemPtr,NULL,args,&w->emSkin.Edge,false);
  if (w->emSkin.Edge.top == -1) w->emSkin.Edge.top = w->emSkin.Edge.left;
  if (w->emSkin.Edge.right == -1) w->emSkin.Edge.right = w->emSkin.Edge.left;
  if (w->emSkin.Edge.bottom == -1) w->emSkin.Edge.bottom = w->emSkin.Edge.top;
  RedrawWnd(w);
}

DEF_BANG(Tile)
{
  if (!w) return;

  CallParser(SkinTile_ConfigItemPtr,NULL,args,&w->emSkin,false);
  RedrawWnd(w);
}

DEF_BANG(Color)
{
  if (!w) return;

  CallParser(NULL,args,&w->cSkin,false);
  CalcColorSet(&w->cSkin);
  RedrawWnd(w);
}
#endif /* NO_SKIN */ 

#ifndef SINGLE_WINDOW
void BangCreate(HWND,LPCSTR args)
{
  WndInfo *w = new WndInfo;
  Config c;
  int res;

  if (!w) return;

  ZeroMemory(w,sizeof(WndInfo));
  ZeroMemory(&c,sizeof(Config));

  res = ParseBoxProc(w,args,(DWORD)&c);

  if (res&PSL_OK)
    AddWindow(w);
  else delete w;
}

void BangDestroy(HWND,LPCSTR args)
{
  WndInfo *w,**pw;
  char one[MAX_LINE_LENGTH];
  int group;

  while (GetToken(args,one,&args,FALSE))
  {
    group = atoi(one);

    pw = &Module.pWnds;
    w = *pw;

    while (w)
    {
      if (group?w->Group==group:match(one,w->Name))
      {
        *pw = w->pNext;
        DestroyWindow(w->hWnd);
        delete w;
      }
      else
      {
        pw = &w->pNext;
      }
      w = *pw;
    }
  }
}
#endif /* SINGLE_WINDOW */ 


// bang command list

lsBangCmdDef Bangs[] = {
  DECL_BANG2(Move)
  DECL_BANG2(Size)
  DECL_BANG2(OnTop)
  DECL_BANG2(Alpha)
  DECL_BANG2(Show)
  DECL_BANG2(LoadBoxHook)
  /* insert new DECL_BANG2 lines here */ 
#ifndef NO_SCRIPT_ACTIONS
  DECL_BANG2(Exec)
  DECL_BANG2(Action)
#endif
#ifndef NO_SKIN
  DECL_BANG2(Bitmap)
  DECL_BANG2(Edge)
  DECL_BANG2(Tile)
  DECL_BANG2(Color)
#endif
#ifndef SINGLE_WINDOW
  DECL_BANG2(Create)
  DECL_BANG2(Destroy)
#endif
	NULL, NULL
};

