#include "module.h"
#include "configstruct.h"
#include "parsers.h"

#define DEBUG_COMPOUND_PARSER

static const char *togglestrings[] = { "toggle", "on", "yes", "true", "off", "no", "false", NULL };
enum togglestringsenum { tseToggle, tseOn, tseYes, tseTrue, tseOff, tseNo, tseFalse };

#define DEF_CIT_PARSER(type) const char * Config##type##Parser(Config *cfg, const char *str, void *base, const ConfigItem *item,bool bInit)
#define DEF_CIT_FREER(type) bool Config##type##Freer(void *base, const ConfigItem *item)

#define PARSER_FAIL NULL
#define PARSER_OK (LPCSTR)1

#define PARSER_RESULT_PTR(name,type) \
  ASSERT_READABLE(item); \
  type *name = reinterpret_cast<type *>(reinterpret_cast<char *>(base)+(item->iOffset)/sizeof(char)); \
  ASSERT_WRITEABLE(name)
#define PARSER_RESULT_VPTR(name) \
  ASSERT_READABLE(item); \
  void *name = reinterpret_cast<void *>(reinterpret_cast<char *>(base)+(item->iOffset)/sizeof(char));
#define PARSER_PARAM(n,name,type) type name = (type)(item->lParam##n)
#define PARSER_DEFAULT(exp) if (bInit) do { exp; } while(0)

#define DEF_CIT(type,subparser)  DEF_CIT_PARSER(type); \
  ConfigItemType CIT_##type = { ConfigItemType::subparser,Config##type##Parser,NULL };
#define DEF_CITF(type,subparser) DEF_CIT_PARSER(type); DEF_CIT_FREER(type); \
  ConfigItemType CIT_##type = { ConfigItemType::subparser,Config##type##Parser,Config##type##Freer };




DEF_CIT(Compound,Compound)
DEF_CITF(CompoundFree,Compound)
DEF_CIT(UnitCompound,Compound)
DEF_CIT(Bool,Simple)
DEF_CIT(Long,Simple)
DEF_CIT(Color,Bracketed)
DEF_CITF(StrToken,Simple)
DEF_CIT(Ordinate,Simple)
DEF_CIT(SimpleOrd,Simple)
DEF_CIT(Flags,Simple)
DEF_CITF(Bitmap,Simple)
DEF_CITF(ActionSet,Simple)
DEF_CIT(Enum,Simple)






void CalcOrd(s_Ordinate *o, int min, int max)
{
  ASSERT_WRITEABLE(o);

  o->Value = (max-min)*o->BasePercent/100 + o->Offset;
  /* if we just add 100 to BasePercent, rounding error is on the wrong side */ 
  if (o->BasePercent<0) o->Value += max;
}

void CalcColor(s_Color *c)
{
  ASSERT_WRITEABLE(c);

  /* get system color */ 
  if (c->SysColor<=COLOR_NOSYSCOLOR)
  {
    c->Value = c->Simple;
  }
  else
  {
    if (c->SysColor>=0)
      c->Value = GetSysColor(c->SysColor);
    /* need to get alpha from simple color */ 
    c->Value = (c->Value&0x00ffffff) | (c->Simple&0xff000000);
  }

  /* scale... */ 
  if (c->Scale)
  {
    int r,g,b;
    r = GetRValue(c->Value);
    g = GetGValue(c->Value);
    b = GetBValue(c->Value);

    r += c->Scale;
    g += c->Scale;
    b += c->Scale;

#define CLAMP(value,min,max) (((value)>=(max))?(max):((value)<=(min)?(min):(value)))
    r = CLAMP(r,0,255);
    g = CLAMP(g,0,255);
    b = CLAMP(b,0,255);

    c->Value = (c->Value&0xff000000) | RGB(r,g,b);
  }
}

bool CalcColorSet(s_EdgeColors *ec)
{
  int i;
  COLORREF v;
  bool bChange = false;

  ASSERT_WRITEABLE(ec);

  v = ec->Face.Value;
  CalcColor(&ec->Face);
  if (v!=ec->Face.Value) bChange = true;

  for (i=0;i<4;++i)
  {
    v = ec->Edge[i].Value;
    if (ec->Edge[i].SysColor==COLOR_SELFFACE)
      ec->Edge[i].Value = ec->Face.Value;
    CalcColor(&ec->Edge[i]);
    if (v!=ec->Edge[i].Value) bChange = true;
  }

  return bChange;
}

const char *IntParser(int *i,const char *s,bool *b)
{
  ASSERT_WRITEABLE(i);
  ASSERT_READABLE(s);
  ASSERT_PTR(b, b==NULL || WRITEABLE(b));

  bool bNeg = false;
  int v = 0;

  if (*s == '+') ++s;
  if (*s == '-') { bNeg = true; ++s; }

  while (*s>='0' && *s<='9')
    v = 10*v + *s++-'0';

  if (bNeg) v = -v;

  if (b) *b = bNeg;
  *i = v;

  return s;
}

bool OrdParser(s_Ordinate *o,const char *s)
{
  bool bNeg; /* -0 is negative */ 
  int v;
  const char *ptr;
  ASSERT_WRITEABLE(o);
  ASSERT_READABLE(s);

  ptr = IntParser(&v,s,&bNeg);
  if (s==ptr) return false;
  s = ptr;
 
  switch (*s)
  {
  case '%':
    switch (*++s)
    {
    case 'r':
    case 'R':
      o->BasePercent += v;
      break;
    default:
      IntParser(&o->Offset,s,NULL);
      o->BasePercent = v;
    }
    break;
  case 'c':
  case 'C':
    o->BasePercent = 50;
    o->Offset = v;
    break;
  case 'r':
  case 'R':
    o->Offset += v;
    break;
  default:
    o->BasePercent = bNeg?100:0;
    o->Offset = v;
  }

  return true;
}

inline int CvtHexDigit(char c)
{
  if (c>='0' && c<='9') return c-'0';
  if (c>='A' && c<='F') return c-'A'+10;
  if (c>='a' && c<='f') return c-'a'+10;
  return -1;
}

bool ColorParser(s_Color *c,const char *s)
{
  int d,v,i;
  bool bOk;

  ASSERT_WRITEABLE(c);
  ASSERT_READABLE(s);

  if (*s==':')
  {
    /* read system color index */ 
    s = IntParser(&v,++s,NULL);
    c->SysColor = v;
  }
  else
  {
    if (*s=='#') ++s;
    v = 0;
    /* read RRGGBBAA */ 
    for (i=0;i<32;i+=4)
    {
      d = CvtHexDigit(*s);
      if (d<0) break;
      v |= d<<i; ++s;
    }
    /* need at least 3 components */ 
    if (i<24) return false;
    /* swap digits, since we read them backwards */ 
    c->SysColor = COLOR_NOSYSCOLOR;
    c->Simple = (v&0x0f0f0f0f)<<4 | (v&0xf0f0f0f0)>>4;
    if (i<32) c->Simple |= 255<<24; /* solid if no alpha */ 
  }

  c->Scale = 0;

  bOk = true;
  do
  {
    switch (*s)
    {
    case '@': /* set alpha */ 
      d = CvtHexDigit(*++s); v = d<<4; if (d<0) break;
      d = CvtHexDigit(*++s); v |= d; if (d<0) break;
      c->Simple = (c->Simple&0x00ffffff) | (v<<24);
      break;
    case '+': /* scale the color */ 
    case '-':
      s = IntParser(&v,s,FALSE);
      c->Scale = v;
      break;
    default:
      bOk = false;
    }
  } while (bOk);

  return true;
}



/* Simple decimal integer. */ 
DEF_CIT_PARSER(Long)
{
  PARSER_RESULT_PTR(res,LONG);
  int v;

  PARSER_PARAM(1,def,LONG);
  PARSER_DEFAULT(*res = def);
  if (str)
  {
    if (str==IntParser(&v,str,NULL)) return PARSER_FAIL;
    *res = v;
  }

  return PARSER_OK;
}

/* Color: (#)RRGGBB(AA) */ 
DEF_CIT_PARSER(Color)
{
  PARSER_RESULT_PTR(res,s_Color);

  PARSER_PARAM(1,def,COLORREF);
  PARSER_PARAM(2,sys,int);
  PARSER_DEFAULT(res->Simple = def; res->SysColor = sys; res->Scale = 0);
  if (str)
  {
    if (ColorParser(res,str))
      return PARSER_OK;
  }
  return PARSER_FAIL;
}

/* SimpleOrd is an integer that can parser ordinal-style strings */ 
DEF_CIT_PARSER(SimpleOrd)
{
  PARSER_RESULT_PTR(res,int);
  s_Ordinate v;

  PARSER_PARAM(1,def,int);
  PARSER_DEFAULT(*res = def);
  /* parse as an ordinal */ 
  if (str && OrdParser(&v,str))
  {
    /* evaluate and store its value */ 
    PARSER_PARAM(2,maxv,int);
    CalcOrd(&v,0,maxv);

    *res = v.Value;
    return PARSER_OK;
  }
  return PARSER_FAIL;
}

/*
  Ordinal: given a range of 0-255:
    50     value =           50
    50c    value = 50%+50 = 157
    50r    value = 50 more than it was before
    50%    value = 50%    = 127
    50%-25 value = 50%-25 = 102
   -50     value = 255-50 = 205
   -50c    value = 50%-50 =  77
   -50%    value = 50%    = 128 (note 50% != -50%)
                etc.
  The value can be recalculated later if the scale changes.
*/
DEF_CIT_PARSER(Ordinate)
{
  PARSER_RESULT_PTR(res,s_Ordinate);

  PARSER_PARAM(1,defp,int);
  PARSER_PARAM(2,defo,int);
  PARSER_DEFAULT(res->Offset=defo; res->BasePercent=defp);
  if (str && OrdParser(res,str))
  {
    TRACE_MSG(("%1 = %2!d!%%+%3!d!", str, res->BasePercent, res->Offset));
    return PARSER_OK;
  }
  return PARSER_FAIL;
}

/* bools accept anything in the togglestrings list, or an empty string */ 
DEF_CIT_PARSER(Bool)
{
  PARSER_RESULT_PTR(res,bool);
  int v;

  PARSER_PARAM(1,def,BOOL);
  PARSER_DEFAULT(*res = def?true:false);

  if (!str || *str==0)
    *res = def?true:false;
  else if ((v = LSStringEnum(str,togglestrings))>=0)
  {
    if (v<tseOn) *res = !*res;
    else if (v<tseOff) *res = def?true:false;
    else *res = def?false:true;
  }
  else
    return PARSER_FAIL;

  return PARSER_OK;
}

/* simple single-token string, must begin with a letter */ 
DEF_CIT_PARSER(StrToken)
{
  PARSER_RESULT_PTR(res,char *);

  if (str && *str>='A')
  {
    *res = new char[lstrlen(str)+1];
    if (!*res) return PARSER_FAIL;
    lstrcpy(*res,str);
    return PARSER_OK;
  }
  return PARSER_FAIL;
}

DEF_CIT_FREER(StrToken)
{
  PARSER_RESULT_PTR(res,char *);

  if (*res)
  {
    delete[] *res;
    *res = NULL;
  }
  return true;
}

/* a set of flags and a number #... */ 
DEF_CIT_PARSER(Flags)
{
  PARSER_RESULT_PTR(group,int);
  bool *res;
  int v = 0;
  bool flags[26]= { false }; /* bucket sort, baby */ 
  PARSER_PARAM(2,f,ConfigItem *);

  PARSER_PARAM(1,defgrp,int);
  PARSER_DEFAULT(*group = defgrp);
  if (str && *str=='#')
  {
    while (*++str)
    {
      /* digit: add to group */ 
      if (*str>='0' && *str<='9')
        v = v*10 + *str-'0';
      /* letter: toggle flag */ 
      else if (*str>='A' && *str<='Z')
        flags[*str-'A'] = !flags[*str-'A'];
      else if (*str>='a' && *str<='z')
        flags[*str-'a'] = !flags[*str-'a'];
    }
    /* note: you cannot set a group 0 with this      */ 
    /* if the default is 0, it will sort of work tho */ 
    if (v) *group = v;

    char line[256];
    while (f->lParam2)
    {
      res = APPLY_OFFSET(base,bool,f->iOffset);
      ASSERT(WRITEABLE(res));
      PARSER_DEFAULT(*res = false);

      /* flag char in struct MUST be upper case! */ 
      ASSERT(f->lParam2>='A' && f->lParam2<='Z');
      if (flags[f->lParam2-'A'])
        *res = !*res;

      if (cfg)
      {
        /* check for separate config item */ 
        ConfigState cs(cfg,f->Key,f->KeyLen);
        if (GetRCString(cfg->key,line,NULL,256))
          f->Type->Parser(cfg,line,base,f,bInit);
      }

      ++f;
    }
    return PARSER_OK;
  }
  return PARSER_FAIL;
}

inline DEF_CIT_PARSER(CompoundFree)
{ return ConfigCompoundParser(cfg,str,base,item,bInit); }

#ifdef DEBUG_COMPOUND_PARSER
#define DCP_MSG(n) TRACE_MSG(n)
#else
#define DCP_MSG(n) (void)0
#endif

DEF_CIT_PARSER(Compound)
{
  PARSER_RESULT_VPTR(res);

  PARSER_PARAM(1,reqsize,int);
//  ASSERT_WRITEABLEC(res,reqsize);

  PARSER_PARAM(2,subitem,ConfigItem *);
  ASSERT_READABLE(subitem);

  char token[MAX_LINE_LENGTH+1];
  char *tptr;
  const char *after = str;
  BOOL bTok = TRUE, bLine = FALSE;
  const char *sParse = PARSER_OK;

  DCP_MSG(("-> Begin compound parser"));

  while (subitem->Type)
  {
    ConfigState cs(cfg,subitem->Key,subitem->KeyLen);

    {
      #ifdef DEBUG_COMPOUND_PARSER
      TRACE_BLOCK(tBlock);
      #endif
      DCP_MSG(("subitem->Key(%1!08X!) = \"%1\"",subitem->Key));

      tptr = token+1;
      if (subitem->Type->Type!=ConfigItemType::Compound
                 /* don't get a token for a compound item      */ 
        && bTok) /* don't get a token if the last try failed   */ 
      {
        bTok = GetToken(str,tptr,&after,
          (subitem->Type->Type==ConfigItemType::Bracketed)?TRUE:FALSE);
        if (bTok)
          DCP_MSG(("GOT TOKEN: \"%1\"", tptr));
        else
          DCP_MSG(("NO TOKEN"));
        /* don't return a NULL pointer */ 
        if (!after)
        {
          after = str + lstrlen(str);
        }
      }
    }
    
    if (cfg)
    {
      #ifdef DEBUG_COMPOUND_PARSER
      TRACE_BLOCK(tBlock);
      #endif
      DCP_MSG(("cfg->key(%1!08X!) = \"%1\"",cfg->key));
      bLine = GetRCLine(cfg->key,token,MAX_LINE_LENGTH,NULL);
      if (bLine)
      {
        --tptr;
        DCP_MSG(("FOUND: \"%1\"", tptr));
      }
      else
        DCP_MSG(("NOT FOUND"));
    }

    if (!bLine && (subitem->Type->Type!=ConfigItemType::Compound && !bTok))
    {
      /* nothing for the parser to read, fail if required    */ 
      if (!subitem->bOptional)
      {
        DCP_MSG(("<- Missing req. param."));
        return PARSER_FAIL;
      }
      /* call parser to set default values                   */ 
      DCP_MSG(("Setting defaults..."));
      sParse = subitem->Type->Parser(cfg,NULL,res,subitem,bInit);
    }
    else
    {
      if (subitem->Type->Type==ConfigItemType::Compound)
      {
        DCP_MSG(("str(%1!08X!) = \"%1\"",str));
        sParse = subitem->Type->Parser(cfg,str,res,subitem,bInit);
        /* if we're reading from config, also try the "separate combined" line */ 
        if (bLine)
        {
          DCP_MSG(("token(%1!08X!) = \"%1\"", tptr));
          sParse = subitem->Type->Parser(cfg,tptr,res,subitem,false);
        }
      }
      else
      {
        DCP_MSG(("token(%1!08X!) = \"%1\"", tptr));
        sParse = subitem->Type->Parser(cfg,tptr,res,subitem,bInit);
      }
    }

    if (sParse)
    {
      /* compound parsers return a pointer to the next token */ 
      if (subitem->Type->Type==ConfigItemType::Compound)
        after = sParse;
      /* otherwise, just skip the token we've used           */ 
      str = after;
    }
    else
    {
      /* parser failed, quit if required                     */ 
      if (!subitem->bOptional)
      {
        DCP_MSG(("<- Req. parser failed."));
        return PARSER_FAIL;
      }
    }

    /* next item */ 
    ++subitem;
    ASSERT_READABLE(subitem);
  }

  DCP_MSG(("<- End compound parser"));

  return str;
}

DEF_CIT_FREER(CompoundFree)
{
  PARSER_RESULT_VPTR(res);
  PARSER_PARAM(2,subitem,ConfigItem *);
  ASSERT_READABLE(subitem);

  while (subitem->Type)
  {
    if (subitem->Type->Freer)
    {
      ASSERT_EXECABLE(subitem->Type->Freer);
      subitem->Type->Freer(res,subitem);
    }
    ++subitem;
  }

  return true;
}

DEF_CIT_PARSER(Bitmap)
{
  PARSER_RESULT_PTR(res,lsEdgeMap);
  HBITMAP hbm;
  int cx,cy;

  ASSERT_WRITEABLE(res);

  PARSER_DEFAULT(res->Bitmap = NULL);

  hbm = LoadLSImage(str,NULL);
  if (hbm)
  {
    if (res->Bitmap) DeleteObject(res->Bitmap);
    res->Bitmap = hbm;
    GetLSBitmapSize(res->Bitmap,&cx,&cy);
    res->Size.cx = cx;
    res->Size.cy = cy;

    const char *p = strrchr(str,'.');
    if (p && stricmp(p,".png")==0)
      /* should make sure it's 32-bit... */ 
      res->HasAlpha = TRUE;
    else
      res->HasAlpha = FALSE;

    return PARSER_OK;
  }

  return PARSER_FAIL;
}

DEF_CIT_FREER(Bitmap)
{
  PARSER_RESULT_PTR(res,HBITMAP);

  if (*res)
  {
    DeleteObject(*res);
    *res = NULL;
  }

  return true;
}

DEF_CIT_PARSER(UnitCompound)
{
  PARSER_RESULT_VPTR(res);
  PARSER_PARAM(1,size,int);
  void *temp;
  
  if (!bInit)
  {
    temp = new BYTE[size];
    CopyMemory(temp,base,size);
  }

  const char *ret = ConfigCompoundParser(cfg,str,base,item,true);

  if (!bInit)
  {
    if (!ret) CopyMemory(res,temp,size);
    delete[] temp;
  }

  return ret;
}

DEF_CIT_PARSER(Enum)
{
  PARSER_RESULT_PTR(res,int);
  PARSER_PARAM(1,def,int);
  PARSER_PARAM(2,ptr,const char **);
  int i;

  PARSER_DEFAULT(*res = def);
  if (str)
  {
    i = LSStringEnum(str,ptr);
    if (i>=0)
    {
      *res = i;
      return PARSER_OK;
    }
  }
  return PARSER_FAIL;
}




DEF_CIT_PARSER(ActionSet)
{
  PARSER_RESULT_PTR(res,ScriptActionSet);
  PARSER_PARAM(1,count,int);
  PARSER_PARAM(2,strings,s_ScriptActionName *);
  ASSERT_READABLEC(strings,count*sizeof(*strings));

  char line[MAX_LINE_LENGTH];

  /* must always be called during config          */ 
  /* bang command should go straight to SetAction */ 
  ASSERT(cfg!=NULL);

  PARSER_DEFAULT(*res = NULL);

  int i=0;
  while (i<count)
  {
    ConfigState cs(cfg,strings[i].Name,strings[i].Length);

    if (GetRCLine(cfg->key,line,MAX_LINE_LENGTH,NULL))
    {
      if (!SetAction(res,count,i,line))
        return PARSER_FAIL;
    }

    ++i;
  }

  return PARSER_OK;
}

DEF_CIT_FREER(ActionSet)
{
  PARSER_RESULT_PTR(res,ScriptActionSet);
  PARSER_PARAM(1,count,int);

  if (!*res) return true;

  ASSERT_WRITEABLEC(*res,count*sizeof(**res));

  int i=0;
  while (i<count)
  {
    if ((*res)[i])
      delete[] (*res)[i];

    ++i;
  }
  delete[] *res;
  *res = NULL;

  return false;
}


